--------------------------------------
-- Namespaces
--------------------------------------
local _, core = ...
core.AltVault = {}

local AltVault = core.AltVault

local NUM_PAGES = 3
local LIST_FILTERS = {
	"All Realms / All Factions",
	"All Realms / Current Faction",
	"All Realms / Opposite Faction",
	"Current Realm / All Factions",
	"Current Realm / Current Faction",
	"Current Realm / Opposite Faction"
}
local SILVER = '|cffc7c7cf%s|r'

StaticPopupDialogs["NO_FILTER_RESULTS"] = {
	text = "No characters match that query.\nTry a different filter.",
	showAlert = true,
    button1 = "OK",
    OnAccept = function()
        StaticPopup_Hide("NO_FILTER_RESULTS")
    end,
    timeout = 0,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3
}

StaticPopupDialogs["NEW_VERSION_RESET"] = {
	text = "You have a new version of AltVault! Your character data has been reset.",
	showAlert = true,
    button1 = "OK",
    OnAccept = function()
        StaticPopup_Hide("NEW_VERSION_RESET")
    end,
    timeout = 0,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3
}

--------------------------------------
-- Defaults (usually a database!)
--------------------------------------
local defaults = {
	theme = {
		r = 0, 
		g = 0.8, -- 204/255
		b = 1,
		hex = "00ccff"
	}
}

--------------------------------------
-- Config functions
--------------------------------------

function AltVault:GetThemeColor()
	local c = defaults.theme
	return c.r, c.g, c.b, c.hex
end

--------------------------------------
-- AltVault Event Handlers
--------------------------------------

function AltVault_OnLoad(self)
	-- Allows close with ESC
	tinsert(UISpecialFrames,"AltVaultFrame");
	AltVault:SetDefaultFilter()
	HybridScrollFrame_SetDoNotHideScrollBar(CharacterJournal.AltListScrollFrame, true)
	CharacterJournal.id = core.AltVault.filteredCharacters[1]
	AltVault:SetupChallengeMaps()
	AltVault:SetupTooltipCounts()
end

function AltVault:SetupTooltipCounts()
	local CLASS_COLOR = '|cff%02x%02x%02x'
	local DEFAULT_COORDS = {0, 1, 0, 1}
	local ALLIANCE_BANNER = 'Interface/Icons/Inv_BannerPvP_02'
	local HORDE_BANNER = 'Interface/Icons/Inv_BannerPvP_01'
	local RACE_TABLE = {
		highmountaintauren = 'highmountain',
		lightforgeddraenei = 'lightforged',
		scourge = 'undead',
		zandalaritroll = 'zandalari',
	}

	function getIcon(character)
		if character.raceKey then
			local race = character.raceKey:lower()
			return format('raceicon-%s-%s', RACE_TABLE[race] or race, character.sex)
		end
	
		return character.faction == 'Alliance' and ALLIANCE_BANNER or HORDE_BANNER, DEFAULT_COORDS
	end
	
	function getIconString(char, size, x, y)
		local icon, coords = getIcon(char.character)
		if coords then
			local u,v,w,z = unpack(coords)
			return CreateTextureMarkup(icon, 128,128, size,size, u,v,w,z, x,y)
		else
			return CreateAtlasMarkup(icon, size,size, x,y)
		end
	end

	local function containerItemCount(container, name)
		local items = container.items
		local containerCount = 0
		if items ~= nil then
			for itemKey, item in pairs(items) do
				local count = 0
				if item and item.searchName == name then
					local stackCount = container.counts[itemKey] or 1
					count = count + stackCount

					if count and count > 0 then
						containerCount = containerCount + count
					else
						-- at least one was found
						containerCount = 1
					end
				end
			end
		end
		return containerCount
	end

	local function equippedItemCount(gear, name)
		local equipCount = 0
		for gearKey, gearItem in pairs(gear) do
			local gearName = GetItemInfo(gearItem)
			if gearName == name then
				equipCount = equipCount + 1
			end
		end
		return equipCount
	end

	local function charCountItemInContainer(charIndex, name, container)
		-- either bags or bank
		local containers = AltVaultDB.characters[charIndex][container]
		local itemCount = 0
		if containers ~= nil then
			for containerKey, containerBag in pairs(containers) do
				local containerCount = containerItemCount(containerBag, name)
				itemCount = itemCount + containerCount
			end
		end
		return itemCount
	end

    local function attachItemTooltip(self)
		local name = self:GetItem()
		local spaced = false
		local total = 0
		local ownerCount = 0
		for key, char in ipairs(AltVaultDB.characters) do
			local bagCount = charCountItemInContainer(key, name, "bags")
			local bankCount = charCountItemInContainer(key, name, "bank")
			local gbankCount = charCountItemInContainer(key, name, "gbank")
			local equipCount = equippedItemCount(char.gear, name)
			local leftText = ""
			local rightText = ""
			if bagCount > 0 or bankCount > 0 or equipCount > 0 or gbankCount > 0 then
				ownerCount = ownerCount + 1
				if not spaced then
					self:AddLine(" ")
					spaced = true
				end
				local r, g, b = GetClassColor(strupper(char.character.classKey))
				local brightness = r + g + b
				local scale = max(1.8 / brightness, 1) * 255
				local classColor = CLASS_COLOR:format(min(r * scale, 255), min(g * scale, 255), min(b * scale, 255)) .. '%s|r'
				leftText = getIconString(char, 14, 0, 0) .. " " .. classColor:format(char.character.name)

				if bagCount > 0 then
					total = total + bagCount
					rightText = rightText .. " Bags:" ..bagCount
				end
				if bankCount > 0 then
					total = total + bankCount
					rightText = rightText .. " Bank:" ..bankCount
				end
				if gbankCount > 0 then
					total = total + gbankCount
					rightText = rightText .. " Vault:" .. gbankCount
				end
				if equipCount > 0 then
					total = total + equipCount
					rightText = rightText .. " Equip:" ..equipCount
				end
				
				self:AddDoubleLine(leftText, SILVER:format(rightText))
				
			end
		end
		if ownerCount > 1 then
			self:AddDoubleLine("Total:", total)
		end

		self:Show()
	end

	local function onTooltipSetItemFunction(tooltip, tooltipData)
		if not AltVaultDB.data.showTooltipCounts then return end
		if (tooltip == GameTooltip or tooltip == ItemRefTooltip) then
			attachItemTooltip(tooltip)
		end
	end

	TooltipDataProcessor.AddTooltipPostCall(Enum.TooltipDataType.Item, onTooltipSetItemFunction)
end

function AltVault_OnShow(self)
	PlaySound(SOUNDKIT.IG_CHARACTER_INFO_OPEN);

	AltVault:SetTotalCharacters()
	-- Set Monies - No mouse interaction
	AltVault:SetTotalMoney()
	-- Set Time Played
	AltVault:SetTotalPlayed()

	CharacterJournal.AltListScrollFrame.update = UpdateAltList
	HybridScrollFrame_CreateButtons(CharacterJournal.AltListScrollFrame, "AltListButtonTemplate")
	UpdateAltList()

	local currentCharacter = AltVaultDB.characters[CharacterJournal.id].character
	local tabContainer = AltInfoTabContainerFrame
	if currentCharacter.bankWarned == false then
		tabContainer.BankTabHelpBox:Show()
	else
		tabContainer.BankTabHelpBox:Hide()
	end

	local tokenPrice = AltVaultDB.data.tokenPrice
	local tokenUpdated = AltVaultDB.data.tokenPriceUpdated
	local tokenWarned = AltVaultDB.data.tokenWarned
	if tokenWarned ~= nil then
		if tokenPrice ~= nil and tokenUpdated ~= nil then
			local tokenAge = AltVault:GetDaysAgo(tokenUpdated)
			if tokenAge < 5 then
				AltVaultTokenIndicator.Icon:SetDesaturated(false)
			else
				AltVaultTokenIndicator.Icon:SetDesaturated(true)
			end
		end
	else
		TokenPriceHelpBox:Show()
		AltVaultDB.data.tokenWarned = true
	end
end

function AltVault:GetDaysAgo(lastDate)
	local currentTime = C_DateAndTime.GetCurrentCalendarTime()
	local updatedDate = time{day=lastDate.monthDay, year=lastDate.year, month=lastDate.month}
	local currentDate = time{day=currentTime.monthDay, year=currentTime.year, month=currentTime.month}
	local daysPassed = difftime(currentDate, updatedDate) / (24 * 60 * 60) -- seconds in a day
	local wholeDays = math.floor(daysPassed)
	return wholeDays
end

function AltVault:IsTodayOrLater(month, day, year)
	local currentCalendarTime = C_DateAndTime.GetCurrentCalendarTime();
	local presentWeekday = currentCalendarTime.weekday;
	local presentMonth = currentCalendarTime.month;
	local presentDay = currentCalendarTime.monthDay;
	local presentYear = currentCalendarTime.year;
	local todayOrLater = false;
	if ( year > presentYear ) then
		todayOrLater = true;
	elseif ( year == presentYear ) then
		if ( month > presentMonth ) then
			todayOrLater = true;
		elseif ( month == presentMonth ) then
			todayOrLater = day >= presentDay;
		end
	end
	return todayOrLater;
end

function TokenIcon_OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOP")
	local tokenPrice = AltVaultDB.data.tokenPrice
	local tokenUpdated = AltVaultDB.data.tokenPriceUpdated
	if tokenPrice ~= nil and tokenUpdated ~= nil then
		local tokenAge = AltVault:GetDaysAgo(tokenUpdated)
		local updatedDay = "today"
		if tokenAge == 1 then
			updatedDay = tokenAge.." day ago"
		elseif tokenAge > 1 then
			updatedDay = tokenAge.." days ago"
		end
		GameTooltip:SetText("WoW Token price updated "..updatedDay)
	else
		GameTooltip:SetText("Visit the Auction House\nto get a WoW Token price")
	end
end

function SetBankWarned(self)
	AltVaultDB.characters[CharacterJournal.id].character.bankWarned = true
	self:GetParent():Hide()
end

function AltVault:SetTotalCharacters()
	CharacterJournal.CharacterCount.Count:SetText(#core.AltVault.filteredCharacters)
end

function AltVault:SetupChallengeMaps()
	-- Set to the stats tab default
	AltVaultStatsFrame.selectedStatsTab = 1;

	-- setup map table
	local mapTable = C_ChallengeMode.GetMapTable()
	for mapIndex = 1, #mapTable do
		local mapFrame = CreateFrame("FRAME", "ChallengeMap"..mapIndex, AltSeasonBest, "AltMPlusSeasonDungeonTile")
		if mapIndex == 1 then
			mapFrame:SetPoint("TOPLEFT", AltSeasonBest, "TOPLEFT", 1, -20)
		else
			local offset = mapIndex - 1
			mapFrame:SetPoint("TOPLEFT", 'ChallengeMap'..offset, "TOPRIGHT", 0, 0)
		end
		local _, _, _, texture = C_ChallengeMode.GetMapUIInfo(mapTable[mapIndex]);
		mapFrame.Icon:SetTexture(texture)
	end
end

function AltVault:SetTotalMoney()
	if AltVaultDB.data.hideGoldValues then
		AltVaultMoneyTotal:Hide()
	else
		AltVaultMoneyTotal:Show()
		local totalMoney = 0
		for charID = 1, #core.AltVault.filteredCharacters do
			local character = AltVaultDB.characters[core.AltVault.filteredCharacters[charID]]
			totalMoney = totalMoney + tonumber(character.character.money)
		end
		AltVault:setMoneyValues(AltVaultMoneyTotal, totalMoney)
	end
	
end

function AltVault:SetTotalPlayed()
	local totalTime = 0
	for charID = 1, #core.AltVault.filteredCharacters do
		local character = AltVaultDB.characters[core.AltVault.filteredCharacters[charID]]
		totalTime = totalTime + tonumber(character.character.played) 
	end
	CharacterJournal.TotalPlayed.Time:SetText(AltVault:Played_TimeBreakDown(totalTime, true))
end

function AltVault:SetDefaultFilter(override)
	if override then
		CharacterJournal.activeFilter = override
	else
		if AltVaultDB.data.rememberFilter and AltVaultDB.data.selectedFilter then
			CharacterJournal.activeFilter = AltVaultDB.data.selectedFilter
		else
			CharacterJournal.activeFilter = 1
		end
	end
	
	AltVault:GetFilteredCharacters()
	CharacterJournal.ListFilterButton:SetText(LIST_FILTERS[CharacterJournal.activeFilter])
end

--------------------------------------
-- Alt List ScrollList Functions
--------------------------------------

function UpdateAltList()
	local scrollFrame = CharacterJournal.AltListScrollFrame
    local buttons = HybridScrollFrame_GetButtons(scrollFrame)
    local offset = HybridScrollFrame_GetOffset(scrollFrame)
    for buttonIndex = 1, #buttons do
        local button = buttons[buttonIndex]
        local itemIndex = buttonIndex + offset

        if itemIndex <= #core.AltVault.filteredCharacters then
			local altInfo = AltVaultDB.characters[core.AltVault.filteredCharacters[itemIndex]]
            local altClass = string.gsub(altInfo.character.class, "%s+", "")
			if altInfo.character.classKey then
				altClass = altInfo.character.classKey
			end
			local iconPath = "Interface\\ICONS\\ClassIcon_"..altClass
			SetPortraitToTexture(button.classIcon, iconPath)
			local _, _, _, colorCode = GetClassColor(strupper(altClass))
			button.altName:SetText(WrapTextInColorCode(altInfo.character.name, colorCode))
			if altInfo.character.faction ~= "Neutral" then
				local factionPath = "Interface\\COMMON\\icon-"..string.lower(altInfo.character.faction)
				button.ListFactionIcon:SetTexture(factionPath)
			end

			if AltVaultDB.data.hideGoldValues then
				button.AltMoneyFrame:Hide()
			else
				button.AltMoneyFrame:Show()
				AltVault:setMoneyValues(button.AltMoneyFrame, altInfo.character.money)
			end

			if altInfo.character.hasMail then button.altHasMailIcon:Show() else button.altHasMailIcon:Hide() end
			if altInfo.character.restState ~= nil and altInfo.character.restState == "Rested" and altInfo.character.restedXP ~= nil and altInfo.character.restedXP > 0 then
				button.altRestedIcon:Show() 
			else 
				button.altRestedIcon:Hide()
			end

			if AltVaultDB.data.tokenPrice ~= nil and altInfo.character.money >= AltVaultDB.data.tokenPrice then
				button.altTokenIcon:Show()
			else
				button.altTokenIcon:Hide()
			end

			if core.AltVault.filteredCharacters[itemIndex] == CharacterJournal.id then
				button.selectedAlt:Show()
				button.classFrame:SetVertexColor(1.0, 0.7, 0.0)
			else
				button.selectedAlt:Hide()
				button.classFrame:SetVertexColor(1.0, 1.0, 1.0)
			end

			button.altServer:SetText(altInfo.character.realm)
			button.altLevel:SetText(altInfo.character.level)
			button:SetID(core.AltVault.filteredCharacters[itemIndex])
            button:Show()
        else
			button:Hide()
			AltVault:setMoneyValues(button.AltMoneyFrame, 0)
        end
	end

    local buttonHeight = buttons[1]:GetHeight()
    local totalHeight = #core.AltVault.filteredCharacters * buttonHeight
    local shownHeight = #buttons * buttonHeight
	HybridScrollFrame_Update(scrollFrame, totalHeight, shownHeight)
end

function AltListItem_OnClick(self)
	if CharacterJournal.id ~= self:GetID() then
		PlaySound(SOUNDKIT.IG_CHAT_SCROLL_UP)
	end 
	UpdateAltView(self:GetID())
end

function UpdateAltView(currentID)
	AltVaultItemsFrame:Hide()
	AltVaultBagsFrame:Hide()
	AltVaultBankFrame:Hide()
	AltVaultProfessionsFrame:Hide()
	CharacterJournal.id = currentID
	UpdateAltList()
	core.AltVaultStorage:AltVaultStorageList(1)
	core.AltVaultCharacter:AltVaultUpdateAltInfo(CharacterJournal.id)
	core.AltVaultCharacter:AltVaultUpdateAltStats(CharacterJournal.id)
	AltVaultItemsFrame:Show()
	AltVaultBagsFrame:Show()
	AltVaultBankFrame:Show()
	AltVaultProfessionsFrame:Show()
end

function AltVaultTab_OnClick(self)
	local tabID = self:GetID()
	local container = AltInfoTabContainerFrame
	PlaySound(SOUNDKIT.IG_CHARACTER_INFO_TAB);
	PanelTemplates_SetTab(AltInfoTabContainerFrame, tabID)
	for pageNum = 1, NUM_PAGES do
		local page = container['page'..pageNum]
		if pageNum == tabID then
			page:Show()
		else
			page:Hide()
		end
	end
end

--------------------------------------
-- Filtering Functions
--------------------------------------

function AltVault:GetFilteredCharacters()
	core.AltVault.filteredCharacters = {}
	local currentID = AltVault:CurrentCharId()
	local currentCharacter = AltVaultDB.characters[currentID]
	local currentFaction = currentCharacter.character.faction
	local currentRealm = currentCharacter.character.realm
	local filter = CharacterJournal.activeFilter
	for id = 1, #AltVaultDB.characters do
		local charData = AltVaultDB.characters[id]
		local isCurrentFaction = charData.character.faction == currentCharacter.character.faction
		local onCurrentRealm = charData.character.realm == currentCharacter.character.realm
		if id ~= currentID then
			if filter == 1 then -- All Realms / All Factions
				table.insert(core.AltVault.filteredCharacters, id)
			elseif filter == 2 and isCurrentFaction then -- All Realms / Current Faction
				table.insert(core.AltVault.filteredCharacters, id)
			elseif filter == 3 and isCurrentFaction ~= true then -- All Realms / Opposite Faction
				table.insert(core.AltVault.filteredCharacters, id)
			elseif filter == 4 and onCurrentRealm then -- Current Realm / All Factions
				table.insert(core.AltVault.filteredCharacters, id)
			elseif filter == 5 and onCurrentRealm and isCurrentFaction then -- Current Realm / Current Faction
				table.insert(core.AltVault.filteredCharacters, id)
			elseif filter == 6 and onCurrentRealm and isCurrentFaction ~= true then -- Current Realm / Opposite Faction
				table.insert(core.AltVault.filteredCharacters, id)
			end
		end
	end
	-- Always have current character at the top of the list unless opposite faction filtered
	if filter ~= 3 and filter ~= 6 then
		table.insert(core.AltVault.filteredCharacters, 1, currentID)
	end
end

function AltVaultListFilterDropDown_OnLoad(self)
	UIDropDownMenu_Initialize(self, AltVaultListFilterDropDown_Initialize, "MENU");
end

function AltVaultListFilterDropDown_Initialize(self, level)
	local info = UIDropDownMenu_CreateInfo();
	for i = 1, #LIST_FILTERS do
		info.text = LIST_FILTERS[i];
		info.checked = i == CharacterJournal.activeFilter;
		info.func = function() AltVaultFilterList(i); end;
		UIDropDownMenu_AddButton(info);
	end
end

function AltVaultFilterList(filterID)
	local previousFilter = CharacterJournal.activeFilter
	CharacterJournal.activeFilter = filterID
	AltVault:GetFilteredCharacters()
	if #core.AltVault.filteredCharacters > 0 then
		AltVaultDB.data.selectedFilter = filterID
		UpdateAltView(core.AltVault.filteredCharacters[1])
		CharacterJournal.ListFilterButton:SetText(LIST_FILTERS[filterID])
		AltVault:SetTotalCharacters()
		AltVault:SetTotalMoney()
		AltVault:SetTotalPlayed()
	else
		CharacterJournal.activeFilter = previousFilter
		AltVault:GetFilteredCharacters()
		StaticPopup_Show("NO_FILTER_RESULTS")
	end
end

function AltVault:ItemInSearch(searchName)
	if searchName == nil or searchName == "" or CharacterJournal.searchText == nil then return false end
	local trimmed = AltVault:TrimString(CharacterJournal.searchText)
	if trimmed == "" then return false end
	local found = string.lower(searchName):find(string.lower(trimmed), 1, true)
	return found ~= nil
end

function AltVault:TrimString(str)
	if str == nil then return end
	return string.gsub(str, '^%s*(.-)%s*$', '%1')
end

--------------------------------------
-- Utility Functions
--------------------------------------

function AltVaultTooltip_Reset(self)
	GameTooltip:Hide()
	if (BattlePetTooltip) then
		BattlePetTooltip:Hide();
	end
	ResetCursor()
end

function AltVault:UpdateViews()
	CharacterJournal:Hide()
	CharacterJournal.id = core.AltVault.filteredCharacters[1]
	CharacterJournal:Show()
end

function AltVault:CurrentCharId()
	local currentId = 1
	currentGUID = UnitGUID("player")
	for key, char in ipairs(AltVaultDB.characters) do
		if char.GUID == currentGUID then
			currentId = key
		end
	end
	return currentId
end

function AltVault:setMoneyValues(moneyContainer, total)
	local _, cButton, sButton, gButton = moneyContainer:GetChildren()
	local gold = floor(total / (COPPER_PER_SILVER * SILVER_PER_GOLD))
	local goldDisplay = BreakUpLargeNumbers(gold)
	local silver = floor((total - (gold * COPPER_PER_SILVER * SILVER_PER_GOLD)) / COPPER_PER_SILVER)
	local copper = mod(total, COPPER_PER_SILVER)

	gButton:SetText(goldDisplay)
	gButton:EnableMouse(false)
	sButton:SetText(silver)
	sButton:EnableMouse(false)
	cButton:SetText(copper)
	cButton:EnableMouse(false)
end

function AltVault:Played_TimeBreakDown(time, long)
	if not time then return 'Unknown' end
	local days = tostring(floor(time / (60 * 60 * 24)))
	local hours = tostring(floor((time - (days * (60 * 60 * 24))) / (60 * 60)))
	local minutes = tostring(floor((time - (days * (60 * 60 * 24)) - (hours * (60 * 60))) / 60))
	local seconds = tostring(mod(time, 60))
	local dayLabel = days == "1" and " day, " or " days, "
	local hourLabel = hours == "1" and " hour, " or " hours, "
	local minuteLabel = minutes == "1" and " minute, " or " minutes, "
	local secondLabel = seconds == "1" and " second" or " seconds"
	if long then
		return days..dayLabel..hours..hourLabel..minutes..minuteLabel..seconds..secondLabel
	else
		return days..dayLabel..hours..":"..minutes..":"..seconds
	end
end